;;;;;;;;;;;;;;;; environments ;;;;;;;;;;;;;;;;

;; environment are linked list of frames
;; a frame holds lists of ids and lists of vals
;; for scoping, each frame has an extra attribute, 
;; the deployment strategy

(define-struct environment ())
(define-struct (empty-env    environment) ())
(define-struct (extended-env environment) (ids vals env ds mark))
;;; mark indicates how the frame came to be:
; 'l slet bindings 
; 'a argument bindings
; 'c call stack propagated bindings
; 'd captured bindings
;;;


;; extend environment with new mappings
(define extend-env
  (lambda (ids vals env ds mark)
    (make-extended-env ids (if (vector? vals) vals 
                               (list->vector vals)) env ds mark)))

;; lookup in env, return #t if found, #f otherwise
(define env-check
  (lambda (id env)
    (env-lookup id env
                (lambda (p e) #t)
                (lambda (id) #f)
                (lambda (e) #t))))

;; lookup in env, only in frames that have the given mark
(define env-check-mark
  (lambda (id env mark)
    (env-lookup id env
                (lambda (p e) #t)
                (lambda (id) #f)
                (lambda (e) (eq? (extended-env-mark e) mark)))))

;; abstract lookup in environment, parametrized by action, escape, and filter
(define env-lookup
  (lambda (id env action esc filter)
    (cond ((empty-env? env) (esc id))
          ((environment? env)
           (let ((pos (find-position id (extended-env-ids env))))
             (if (and (number? pos) (filter env))
                 (action pos env) 
                 (env-lookup id (extended-env-env env) action esc filter))))
          (else (error 'env-lookup-esc-filter "not an environment -- ~s" env)))))

(define env-get
  (lambda (id env filter)
    (env-lookup id env 
                (lambda (pos env) (vector-ref (extended-env-vals env) pos))
                (lambda (id) (error 'env-lookup "No binding for ~s" id))
                filter)))

(define env-set!
  (lambda (id env nval)
    (env-lookup id env
                (lambda (pos env) (vector-set! (extended-env-vals env) pos nval))
                (lambda (id) (error 'env-lookup "No binding for ~s" id))
                (lambda (e) #t))))

(define (print-env env)
  (if (empty-env? env) (printf "/") 
      (begin (printf "~a[~a]->" (extended-env-ids env) (extended-env-mark env))
             (print-env (extended-env-env env)))))
  

;; HELPER FUNCTIONS

(define find-if
  (lambda (pred lst)
    (cond ((null? lst)      #f)
	  ((pred (car lst)) (car lst))
	  (else             (find-if pred (cdr lst))))))

(define collect-if
  (lambda (pred lst)
    (cond ((null? lst)      '())
	  ((pred (car lst)) (cons (car lst) (collect-if pred (cdr lst))))
	  (else             (collect-if pred (cdr lst))))))

;; returns an environment made of frames in env that match env (same order)
; lifted to take mark parameter
(define collect-env-if
  (lambda (mark)
    (lambda (pred env)
      (cond ((empty-env? env) (make-empty-env))
            ((pred env)       (extend-env (extended-env-ids env) 
                                          (extended-env-vals env)
                                          ((collect-env-if mark) pred (extended-env-env env))
                                          (extended-env-ds env)
                                          (if (not mark) (extended-env-mark env) mark)))
            (else             ((collect-env-if mark) pred (extended-env-env env)))))))

;; make bottom env extend top env
(define env-append
  (lambda (bottom top)
    (cond ((empty-env? bottom) top)
          (else (extend-env (extended-env-ids bottom) 
                            (extended-env-vals bottom)
                            (env-append (extended-env-env bottom) top)
                            (extended-env-ds bottom)
                            (extended-env-mark bottom))))))

(define find-position
  (lambda (sym lst)
    (let loop ((tail lst) (pos 0))
      (cond ((null? tail) #f)
	    ((eqv? sym (car tail)) pos)
	    (else (loop (cdr tail) (+ pos 1)))))))

(define find-last-position
  (lambda (sym lst)
    (let loop
	((tail lst) (curpos 0) (lastpos #f))
      (cond
       ((null? tail) lastpos)
       ((eqv? sym (car tail))
	(loop (cdr tail) (+ curpos 1) curpos))
       (else (loop (cdr tail) (+ curpos 1) lastpos))))))

