(load "scope.scm")
(load "test.scm")

;; basic tests
(test (run '10) 10)
(test/pred (run '(fun (x y) (+ x y))) closure?)
(test (run '(let ((f (fun (x y) (+ x y))))
              (- (f 3 4)(f 1 1)))) 5)

; static scoping preserved
(test (run '(let ((y 10))
              (let ((foo (fun (x) (+ x y))))
                (let ((y 15))
                  (foo 2)))))
      12)

; dynamic scoping
;; a- initial binding in scope 
(test (run '(dlet ((y 10))
                  (let ((foo (fun (x) (+ x y))))
                    (foo 2))))
      12)

;; b- new binding in scope 
(test (run '(let ((foo (dlet ((y 10))
                             (fun (x) (+ x y)))))
              (dlet ((y 3))
                    (foo 2))))
      5)

;; c- newer binding in scope 
(test (run '(dlet ((y 10))
                  (let ((foo (fun (x) (+ x y))))
                    (dlet ((y 15))
                          (foo 2)))))
      17)

;; static scoping has precedence
; (more precisely, captured environment has precedence over current environment)
(test (run '(let ((y 10))
              (let ((foo (fun (x) (+ x y))))
                (dlet ((y 15))
                      (foo 2)))))
      12)

;; dlet is not sticky
(test/str (run '(dlet ((output "default output"))
                      (let ((worker (fun () 
                                         (write-string output)
                                         (fun ()
                                              (write-string output)))))
                        (let ((worker2 (dlet ((output "this one "))
                                             (worker))))
                          (worker2)))))
          "this one default output")
      
 
;; use ds-all to have fluid and sticky
(test/str (run '(dlet ((output "default output"))
                      (let ((worker (fun () 
                                         (write-string output)
                                         (fun ()
                                              (write-string output)))))
                        (let ((worker2 (plet ((output "this one "))
                                             (worker))))
                          (worker2)))))
          "this one this one ")


;; flat let, binding only available directly in the body -- not in nested lambdas
(test/exn (run '(flet ((x 4) (y 3))
                      (+ y 1) ;; ok 
                      ((fun () (write x))))) ;;exn!
          "env-lookup: No binding for x")


;; if light-on, use def-time binding
;; otherwise take the most recent one
(test/str (run '(slet (list (fun (jp) #f) (fun (jp) #t) (fun (jp) (context-get))) 
                      ((output "with light "))
                      (let ((worker (fun ()
                                         (write-string output))))
                        
                        ;;light is on, ignore dynamic binding
                        (context-set! #t)
                        (dlet ((output "ignore this one"))
                              (worker))
                        ;; light is off, take dynamic binding
                        (context-set! #f)
                        (dlet ((output "take this one"))
                              (worker)))))
          "with light take this one")


;; use def-time binding unless there is a dynamic binding available
(test/str (run '(let ((single-binding (fun (id)
                                           (fun (jp env)
                                                (let ((env (env-find id env)))
                                                  (if env
                                                      (not (env-check id (env-next env)))
                                                      #t))))))
                  
                  (let ((worker  (slet (list (fun (jp) #f) (fun (jp) #t) (single-binding 'output))
                                       ((output "def time "))
                                       (fun (x)
                                            (write-string output)))))
                    
                    ;; no dynamic binding for output, use def-time
                    (worker 1)
                    ;; take dynamic binding
                    (dlet ((output "take this one"))
                          (worker 1)))))
          "def time take this one")




            
;; same with soft-let
;; use def-time binding unless there is a dynamic binding available
(test/str (run '(let ((worker  (soft-let (output "def time ")
                                         (fun (x) (write-string output)))))
                  ;; no dynamic binding for output, use def-time
                  (worker 1)
                  ;; take dynamic binding
                  (dlet ((output "take this one"))
                        (worker 1))))
          "def time take this one")



;;-- EXCEPTIONS 
;;note: handler is naive/incorrect, should use continuation captured at dlet time

;; exception handlers, nesting
(test/str (run 
           '(dlet ((handler (fun (e) (write-string "original handler-"))))
                  (let ((prog-with-exc  (fun () (handler "not good-"))))
                    (+ 1 (prog-with-exc))
                    (dlet ((handler (fun (e) (write-string "new handler-"))))
                          (prog-with-exc)))))
          "original handler-new handler-")

;; function escapes original handler
;; applied in a context with a handler defined.
(test/str (run 
           '(let ((get-fun (fun () (fun () (handler "not good")))))
              
              (let ((the-fun (dlet ((handler (fun (e) (write-string "original handler-"))))
                                   (get-fun))))
                (dlet ((handler (fun (e) (write-string "new handler-"))))
                      (the-fun)))))
          "new handler-")
          

;; function escapes original handler
;; no handler defined
(test/exn (run 
           '(let ((get-fun (fun () (fun () (handler "not good")))))
              
              (let ((the-fun (dlet ((handler (fun (e) (write-string "original handler-"))))
                                   (get-fun))))
                (the-fun))))
          "env-lookup: No binding for handler")

;; function escapes original handler
;; no handler defined -- uses plet so works
(test/str (run 
           '(let ((get-fun (fun () (fun () (handler "not good")))))
              
              (let ((the-fun (plet ((handler (fun (e) (write-string "handler-"))))
                                   (get-fun))))
                (the-fun))))
          "handler-")

;; function escapes original handler
;; new handler defined -- but uses plet so does not use "most recent handler"
(test/str (run 
           '(let ((get-fun (fun () (fun () (handler "not good")))))
              
              (let ((the-fun (plet ((handler (fun (e) (write-string "handler-"))))
                                   (get-fun))))
                (dlet ((handler (fun (e) (write-string "new handler-"))))
                      (the-fun)))))
          "handler-")

;; function escapes original handler
;; no handler defined -- uses soft-plet so works
(test/str (run 
           '(let ((get-fun (fun () (fun () (handler "not good")))))
              
              (let ((the-fun (soft-plet (handler (fun (e) (write-string "handler-")))
                                        (get-fun))))
                (the-fun))))
          "handler-")

;; uses most recent dynamic binding (soft-plet behaves as dlet when no closures are involved)
(test/str (run 
           '(dlet ((handler (fun (e) (write-string "handler-1"))))
                    (soft-plet (handler (fun (e) (write-string "new handler-")))
                               (handler "not good"))))
          "new handler-")

;; function escapes original handler
;; new handler defined -- uses soft-plet so uses the "most recent handler"
(test/str (run 
           '(let ((get-fun (fun () (fun () (handler "not good")))))
              
              (let ((the-fun (soft-plet (handler (fun (e) (write-string "handler-")))
                                        (get-fun))))
                (dlet ((handler (fun (e) (write-string "new handler-"))))
                      (the-fun)))))
          "new handler-")

